from time import sleep
from x_ray.healthcheck.check_items.base_item import BaseItem
from x_ray.healthcheck.rules.cache_rule import CacheRule
from x_ray.healthcheck.rules.connections_rule import ConnectionsRule
from x_ray.healthcheck.rules.query_targeting_rule import QueryTargetingRule
from x_ray.healthcheck.shared import MAX_MONGOS_PING_LATENCY, discover_nodes, enum_all_nodes, enum_result_items
from x_ray.utils import format_size, escape_markdown, green, yellow

SERVER_STATUS_INTERVAL = 5


class ServerStatusItem(BaseItem):
    def __init__(self, output_folder, config=None):
        super().__init__(output_folder, config)
        self._name = "Server Status Information"
        self._description = "Collects and reviews server status metrics.\n\n"
        self._description += "- Whether used/total connection ratio is too high.\n"
        self._description += "- Whether query targeting is poor.\n"
        self._description += "- Whether the cache read into rate is too high.\n"
        self._description += "- Whether updates ratio is too high.\n"
        self._description += "- Whether dirty data ratio is too high.\n"
        self._description += "- Whether cache fill ratio is too high.\n"
        self._query_targeting_rule = QueryTargetingRule(config)
        self._connections_rule = ConnectionsRule(config)
        self._cache_rule = CacheRule(config)

    def test(self, *args, **kwargs):
        """
        Run the server status test.
        """
        client = kwargs.get("client")
        parsed_uri = kwargs.get("parsed_uri")

        def func_first_req(set_name, node, server_status):
            # First request do nothing but return the server status as is.
            # The test will be in the 2nd request.
            return [], {"server_status": server_status}

        def func_2nd_req(set_name, node, server_status):
            host = node["host"]
            test_result1, raw_result1 = self._connections_rule.apply(server_status, extra_info={"host": host})
            test_result2, raw_result2 = (
                self._query_targeting_rule.apply(server_status, extra_info={"host": host})
                if set_name != "mongos"
                else ([], None)
            )
            test_result = test_result1 + test_result2
            self.append_test_results(test_result)
            raw_result = {"connections": raw_result1, "query_targeting": raw_result2, "server_status": server_status}

            return test_result, raw_result

        def enumerator(set_name, node, **kwargs):
            host = node["host"]
            if "pingLatencySec" in node and node["pingLatencySec"] > MAX_MONGOS_PING_LATENCY:
                self._logger.warning(
                    yellow(
                        f"Skip {host} because it has been irresponsive for {node['pingLatencySec'] / 60:.2f} minutes."
                    )
                )
                return None, None
            client = node["client"]
            server_status = client.admin.command("serverStatus")
            func_req = kwargs.get("func_req")
            test_result, raw_result = func_req(set_name, node, server_status)
            return test_result, raw_result

        nodes = discover_nodes(client, parsed_uri)

        def func_all_first(set_name, node, **kwargs):
            return enumerator(set_name, node, func_req=func_first_req, **kwargs)

        result1 = enum_all_nodes(
            nodes,
            func_mongos_member=func_all_first,
            func_rs_member=func_all_first,
            func_shard_member=func_all_first,
            func_config_member=func_all_first,
        )
        # Sleep for 5s to capture next status.
        self._logger.info("Sleep %s to capture next server status.", green(f"{SERVER_STATUS_INTERVAL} seconds"))
        sleep(SERVER_STATUS_INTERVAL)

        def func_all_2nd(set_name, node, **kwargs):
            return enumerator(set_name, node, func_req=func_2nd_req, **kwargs)

        result2 = enum_all_nodes(
            nodes,
            func_mongos_member=func_all_2nd,
            func_rs_member=func_all_2nd,
            func_shard_member=func_all_2nd,
            func_config_member=func_all_2nd,
        )

        # These metrics needs to compare 2 `serverStatus` results
        cache = {}

        def func_data_member(set_name, node, **kwargs):
            raw_result = node.get("rawResult", {})
            host = node["host"]
            if not raw_result:
                cache[host] = {
                    "setName": set_name,
                    "host": host,
                    "cacheSize": "n/a",
                    "inCacheSize": "n/a",
                    "readInto": "n/a",
                    "writtenFrom": "n/a",
                }
                return

            if host not in cache:
                # Enumerating result1
                cache[host] = raw_result
            else:
                # Enumerating result2
                test_result, parsed_data = self._cache_rule.apply(
                    raw_result["server_status"],
                    extra_info={"host": host, "server_status": cache[host]["server_status"]},
                )
                self.append_test_results(test_result)
                # Attach the test result and raw result to the original result.
                node["testResult"].extend(test_result)
                node["rawResult"]["cache"] = parsed_data

        enum_result_items(
            result1,
            func_rs_member=func_data_member,
            func_shard_member=func_data_member,
            func_config_member=func_data_member,
        )
        enum_result_items(
            result2,
            func_rs_member=func_data_member,
            func_shard_member=func_data_member,
            func_config_member=func_data_member,
        )

        op_counters = {}

        def func_all_member(set_name, node, **kwargs):
            raw_result = node.get("rawResult", {})
            host = node["host"]
            if not raw_result:
                op_counters[host] = {
                    "set_name": set_name,
                    "host": host,
                    "insert": "n/a",
                    "query": "n/a",
                    "update": "n/a",
                    "delete": "n/a",
                    "command": "n/a",
                    "getmore": "n/a",
                }
                return
            ops = raw_result["server_status"]["opcounters"]
            if host not in op_counters:
                op_counters[host] = {
                    "set_name": set_name,
                    "host": host,
                    "insert": ops.get("insert", 0),
                    "query": ops.get("query", 0),
                    "update": ops.get("update", 0),
                    "delete": ops.get("delete", 0),
                    "command": ops.get("command", 0),
                    "getmore": ops.get("getmore", 0),
                }
            else:
                inserts = ops["insert"]
                reads = ops["query"]
                updates = ops["update"]
                deletes = ops["delete"]
                commands = ops["command"]
                getmores = ops["getmore"]
                op_counters[host] = {
                    "set_name": set_name,
                    "host": host,
                    "insert": inserts - op_counters[host]["insert"],
                    "query": reads - op_counters[host]["query"],
                    "update": updates - op_counters[host]["update"],
                    "delete": deletes - op_counters[host]["delete"],
                    "command": commands - op_counters[host]["command"],
                    "getmore": getmores - op_counters[host]["getmore"],
                }
                node["rawResult"]["op_counters"] = op_counters[host]

        enum_result_items(
            result1,
            func_mongos_member=func_all_member,
            func_rs_member=func_all_member,
            func_shard_member=func_all_member,
            func_config_member=func_all_member,
        )
        enum_result_items(
            result2,
            func_mongos_member=func_all_member,
            func_rs_member=func_all_member,
            func_shard_member=func_all_member,
            func_config_member=func_all_member,
        )
        self.captured_sample = [result1, result2]

    @property
    def review_result(self):
        result = self.captured_sample
        _, result2 = result
        data = []
        conn_table = {
            "type": "table",
            "caption": "Connections",
            "notes": "- `Rejected` is only available for MongoDB 6.3 and later.\n"
            + "- `Threaded` is only available for MongoDB 5.0 and later.\n",
            "columns": [
                {"name": "Component", "type": "string"},
                {"name": "Host", "type": "string"},
                {"name": "Current", "type": "decimal"},
                {"name": "Available", "type": "decimal"},
                {"name": "Active", "type": "decimal"},
                {"name": "Created", "type": "decimal"},
                {"name": "Rejected", "type": "decimal"},
                {"name": "Threaded", "type": "decimal"},
            ],
            "rows": [],
        }
        current = []
        active = []
        data_conn = {}
        opcounters_table = {
            "type": "table",
            "caption": "Operation Counters",
            "columns": [
                {"name": "Component", "type": "string"},
                {"name": "Host", "type": "string"},
                {"name": "Inserts", "type": "decimal"},
                {"name": "Queries", "type": "decimal"},
                {"name": "Updates", "type": "decimal"},
                {"name": "Deletes", "type": "decimal"},
                {"name": "Commands", "type": "decimal"},
                {"name": "Getmores", "type": "decimal"},
            ],
            "rows": [],
        }
        inserts = []
        queries = []
        updates = []
        deletes = []
        commands = []
        getmores = []
        data_ops = {}
        data.append(conn_table)
        data.append({"type": "chart", "data": data_conn})
        data.append(opcounters_table)
        data.append({"type": "chart", "data": data_ops})

        def func_all_members(set_name, node, **kwargs):
            raw_result = node.get("rawResult", {})
            if not raw_result:
                conn_table["rows"].append(
                    [escape_markdown(set_name), node["host"], "n/a", "n/a", "n/a", "n/a", "n/a", "n/a"]
                )
                opcounters_table["rows"].append(
                    [escape_markdown(set_name), node["host"], "n/a", "n/a", "n/a", "n/a", "n/a", "n/a"]
                )
                data_conn[f"{set_name}/{node['host']}"] = {
                    "current": 0,
                    "available": 0,
                    "active": 0,
                    "totalCreated": 0,
                    "rejected": 0,
                    "threaded": 0,
                }
                data_ops[f"{set_name}/{node['host']}"] = {
                    "insert": 0,
                    "query": 0,
                    "update": 0,
                    "delete": 0,
                    "command": 0,
                    "getmore": 0,
                }
                return
            host = node["host"]
            conns = raw_result.get("connections", {})
            conn_table["rows"].append(
                [
                    escape_markdown(set_name),
                    host,
                    conns.get("current", 0),
                    conns.get("available", 0),
                    conns.get("active", 0),
                    conns.get("totalCreated", 0),
                    conns.get("rejected", "n/a"),
                    conns.get("threaded", "n/a"),
                ]
            )
            data_conn[f"{set_name}/{host}"] = {
                "current": conns.get("current", 0),
                "available": conns.get("available", 0),
                "active": conns.get("active", 0),
                "totalCreated": conns.get("totalCreated", 0),
                "rejected": conns.get("rejected", 0),
                "threaded": conns.get("threaded", 0),
            }
            active.append(conns.get("active", 0))
            current.append(conns.get("current", 0) - conns.get("active", 0))
            opcounters = raw_result.get("op_counters", {})
            opcounters_table["rows"].append(
                [
                    escape_markdown(set_name),
                    host,
                    opcounters.get("insert", 0),
                    opcounters.get("query", 0),
                    opcounters.get("update", 0),
                    opcounters.get("delete", 0),
                    opcounters.get("command", 0),
                    opcounters.get("getmore", 0),
                ]
            )
            data_ops[f"{set_name}/{host}"] = {
                "insert": opcounters.get("insert", 0),
                "query": opcounters.get("query", 0),
                "update": opcounters.get("update", 0),
                "delete": opcounters.get("delete", 0),
                "command": opcounters.get("command", 0),
                "getmore": opcounters.get("getmore", 0),
            }
            inserts.append(opcounters.get("insert", 0))
            queries.append(opcounters.get("query", 0))
            updates.append(opcounters.get("update", 0))
            deletes.append(opcounters.get("delete", 0))
            commands.append(opcounters.get("command", 0))
            getmores.append(opcounters.get("getmore", 0))

        enum_result_items(
            result2,
            func_mongos_member=func_all_members,
            func_rs_member=func_all_members,
            func_shard_member=func_all_members,
            func_config_member=func_all_members,
        )
        qt_table = {
            "type": "table",
            "caption": "Query Targeting",
            "columns": [
                {"name": "Component", "type": "string"},
                {"name": "Host", "type": "string"},
                {"name": "Scanned / Returned", "type": "decimal"},
                {"name": "Scanned Objects / Returned", "type": "decimal"},
            ],
            "rows": [],
        }
        cache_table = {
            "type": "table",
            "caption": "WiredTiger Cache",
            "columns": [
                {"name": "Component", "type": "string"},
                {"name": "Host", "type": "string"},
                {"name": "Cache Size", "type": "decimal"},
                {"name": "In-Cache Size", "type": "decimal"},
                {"name": "Read Into", "type": "decimal"},
                {"name": "Written From", "type": "decimal"},
            ],
            "rows": [],
        }
        cache_sizes = []
        in_cache_sizes = []
        read_into_sizes = []
        written_from_sizes = []
        data_cache = {}
        data.append(cache_table)
        data.append({"type": "chart", "data": data_cache})
        data.append(qt_table)

        def func_data_member(set_name, node, **kwargs):
            raw_result = node.get("rawResult", {})
            host = node["host"]
            if not raw_result:
                cache_table["rows"].append([escape_markdown(set_name), host, "n/a", "n/a", "n/a", "n/a"])
                qt_table["rows"].append([escape_markdown(set_name), node["host"], "n/a", "n/a"])
                data_cache[f"{set_name}/{host}"] = {
                    "cacheSize": 0,
                    "inCacheSize": 0,
                    "readInto": 0,
                    "forUpdates": 0,
                    "dirty": 0,
                    "writtenFrom": 0,
                }
                return
            cache = raw_result.get("cache", {})
            cache_table["rows"].append(
                [
                    escape_markdown(set_name),
                    escape_markdown(host),
                    format_size(cache.get("cacheSize", 0)),
                    format_size(cache.get("inCacheSize", 0)),
                    f"{format_size(cache.get('readInto', 0))}/s",
                    f"{format_size(cache.get('writtenFrom', 0))}/s",
                ]
            )
            data_cache[f"{set_name}/{host}"] = {
                "cacheSize": cache.get("cacheSize", 0),
                "inCacheSize": cache.get("inCacheSize", 0),
                "readInto": cache.get("readInto", 0),
                "forUpdates": cache.get("forUpdates", 0),
                "dirty": cache.get("dirty", 0),
                "writtenFrom": cache.get("writtenFrom", 0),
            }
            cache_sizes.append(cache.get("cacheSize", 0))
            in_cache_sizes.append(cache.get("inCacheSize", 0))
            read_into_sizes.append(cache.get("readInto", 0))
            written_from_sizes.append(cache.get("writtenFrom", 0))
            query_targeting = raw_result.get("query_targeting", {})
            qt_table["rows"].append(
                [
                    escape_markdown(set_name),
                    host,
                    f"{query_targeting.get('scanned/returned', 0):.2f}",
                    f"{query_targeting.get('scanned_objects/returned', 0):.2f}",
                ]
            )

        enum_result_items(
            result2,
            func_rs_member=func_data_member,
            func_shard_member=func_data_member,
            func_config_member=func_data_member,
        )

        return {"name": self.name, "description": self.description, "data": data}



from x_ray.log_analysis.log_items.base_item import BaseItem
from x_ray.log_analysis.shared import to_json


class InfoItem(BaseItem):
    def __init__(self, output_folder, config):
        super().__init__(output_folder, config)
        self.name = "Basic Info"
        self.description = "Basic information about the instance."
        self._cache = {}

        self._ids = [
            20721,  # Process Details
            20722,  # Node is a member of a replica set
            5853300,  # current featureCompatibilityVersion value
            23403,  # Build Info
            51765,  # Operating System
            21951,  # Options set by command line
            4913010,  # Certificate information
            4615611,  # MongoDB starting
        ]

    def analyze(self, log_line):
        log_id = log_line.get("id", "")
        index = self._ids.index(log_id) if log_id in self._ids else -1
        attr = log_line.get("attr", {})
        if index in [0, 7]:
            # Process Details
            self._process_details(attr)
        elif index == 1:
            # Node is a member of a replica set
            self._process_replica_set(attr)
        elif index == 2:
            # current featureCompatibilityVersion value
            self._process_feature_compatibility(attr)
        elif index == 3:
            # Build Info
            self._process_build_info(attr)
        elif index == 4:
            # Operating System
            self._process_operating_system(attr)
        elif index == 5:
            # Options set by command line
            self._process_command_line_options(attr)
        elif index == 6:
            # Certificate information
            self._process_certificate_info(attr)

    def _process_details(self, attr):
        self._cache["process"] = {
            "pid": attr.get("pid", "Unknown"),
            "host": attr.get("host", "Unknown"),
            "port": attr.get("port", "Unknown"),
        }

    def _process_replica_set(self, attr):
        self._cache["replica_set"] = attr

    def _process_feature_compatibility(self, attr):
        self._cache["fcv"] = attr.get("featureCompatibilityVersion", "Unknown")

    def _process_build_info(self, attr):
        build_info = attr.get("buildInfo", {})
        self._cache["build_info"] = {
            "version": build_info.get("version", "Unknown"),
            "modules": build_info.get("modules", []),
            "environment": build_info.get("environment", {}),
        }

    def _process_operating_system(self, attr):
        os_info = attr.get("os", {})
        self._cache["os"] = {
            "name": os_info.get("name", "Unknown"),
            "version": os_info.get("version", "Unknown"),
        }

    def _process_command_line_options(self, attr):
        options = attr.get("options", {})
        self._cache["command_line_options"] = options

    def _process_certificate_info(self, attr):
        cert_info = attr
        self._cache["cert_info"] = cert_info

    def review_results_markdown(self, f):
        super().review_results_markdown(f)
        if not self._cache:
            f.write("No basic info found in the logs.\n")
            return
        process = self._cache.get("process", None)
        build_info = self._cache.get("build_info", None)
        fcv = self._cache.get("fcv", None)
        if build_info:
            f.write("### Process Info\n\n")
            version = build_info.get("version", "Unknown")
            if "enterprise" in build_info.get("modules", []):
                version += "-ent"
            f.write(f"- MongoDB `{version}` ")
            if fcv:
                f.write(f" (FCV: `{fcv}`)")
            if process:
                f.write(
                    f" PID `{process.get('pid', 'Unknown')}` running on `{process.get('host', 'Unknown')}:{process.get('port', 'Unknown')}`\n"
                )
            f.write("\n")
        cert_info = self._cache.get("cert_info", None)
        if cert_info:
            f.write("### Certificate Info\n\n")
            key_file = cert_info.get("keyFile", "Unknown")
            subject = cert_info.get("subject", "Unknown")
            issuer = cert_info.get("issuer", "Unknown")
            valid_from = cert_info.get("notValidBefore", "Unknown")
            valid_to = cert_info.get("notValidAfter", "Unknown")
            cert_type = cert_info.get("type", "Unknown")
            f.write(f"- Key File: `{key_file}`\n")
            f.write(f"- Type: `{cert_type}`\n")
            f.write(f"- Subject: `{subject}`\n")
            f.write(f"- Issuer: `{issuer}`\n")
            f.write(f"- Valid: `{valid_from}` ~ `{valid_to}`\n\n")
        guest_os = self._cache.get("os", None)
        if guest_os:
            f.write("### Operating System\n\n")
            f.write(f"- {guest_os.get('name', 'Unknown')}\n")
            f.write(f"- {guest_os.get('version', 'Unknown')}\n\n")
        replica_set = self._cache.get("replica_set", {})
        rs_config = replica_set.get("config", None)
        if replica_set:
            f.write("### Replica Set Config\n\n")
            f.write(
                f"Replica Set Name: `{rs_config.get('_id', 'Unknown')}`, member state: `{replica_set.get('memberState', 'Unknown')}`\n\n"
            )
            f.write("|Member|Host|Arbiter|Priority|Votes|Hidden|Delay|\n")
            f.write("|------|----|-------|--------|-----|------|-----|\n")
            for member in rs_config.get("members", []):
                f.write(
                    f"|{member.get('_id', 'Unknown')}|{member.get('host', 'Unknown')}|{member.get('arbiterOnly', False)}|{member.get('priority', 0)}|{member.get('votes', 0)}|{member.get('hidden', False)}|{member.get('secondaryDelaySecs', 0)}|\n"
                )
            f.write("\n")
        command_line_options = self._cache.get("command_line_options", None)
        if command_line_options:
            f.write("### Command Line Options\n\n")
            f.write('<div id="cmd_options">\n')
            f.write("```json\n")
            f.write(to_json(command_line_options, indent=4))
            f.write("\n```\n\n")
            f.write("</div>\n")



"""Analyze slow query rates from log lines."""

from datetime import datetime
import math
from x_ray.log_analysis.log_items.base_item import BaseItem


class SlowRateItem(BaseItem):
    def __init__(self, output_folder: str, config):
        super().__init__(output_folder, config, show_reset=True)
        self._cache = None
        self.name = "Slow Rate"
        self.description = "Analyse the rate of slow queries."

    def analyze(self, log_line):
        log_id = log_line.get("id", "")
        if log_id != 51803:  # Slow query
            return
        time = log_line.get("t")
        ts = math.floor(time.timestamp())
        time_min = datetime.fromtimestamp(ts - (ts % 60))

        if self._cache is None or self._cache["time"] != time_min:
            if self._cache is not None:
                # New minute, write previous minute's data
                self._write_output()
            # First time or new minute bucket
            self._cache = {"time": time_min, "total_slow_ms": 0, "count": 0, "byNs": {}}

        attr = log_line.get("attr", {})
        slow_ms = attr.get("durationMillis", 0)
        ns = attr.get("ns", "unknown")
        self._cache["count"] += 1
        self._cache["total_slow_ms"] += slow_ms
        if ns not in self._cache["byNs"]:
            self._cache["byNs"][ns] = {"count": 0, "total_slow_ms": 0}
        self._cache["byNs"][ns]["count"] += 1
        self._cache["byNs"][ns]["total_slow_ms"] += slow_ms

    def review_results_markdown(self, f):
        super().review_results_markdown(f)
        f.write(f'<canvas id="canvas_{self.__class__.__name__}" width="400" height="200"></canvas>\n')
        f.write(f'<div class="pie"><canvas id="canvas_{self.__class__.__name__}_byns" height="200"></canvas></div>\n')
        f.write(
            f'<div class="pie"><canvas id="canvas_{self.__class__.__name__}_byns_ms" height="200"></canvas></div>\n'
        )



"""
Analyze MongoDB query patterns from log entries.
"""

from x_ray.utils import json_hash


DATA_TYPES = [
    "$binary",
    "$date",
    "$numberLong",
    "$numberInt",
    "$numberDecimal",
    "$oid",
    "$timestamp",
]
QUERY_OPERATORS = [
    "$all",
    "$size",
    "$elemMatch",  # Array operators
    "$bitsAllClear",
    "$bitsAllSet",
    "$bitsAnyClear",
    "$bitsAnySet",  # Bitwise operators
    "$gt",
    "$gte",
    "$lt",
    "$lte",
    "$eq",
    "$ne",
    "$in",
    "$nin",  # Comparison operators
    "$exists",
    "$type",  # Data Type operators
    "$box",
    "$center",
    "$centerSphere",
    "$geoIntersects",
    "$geometry",
    "$geoWithin",
    "$maxDistance",
    "$minDistance",
    "$near",
    "$nearSphere",
    "$polygon",  # Geospatial operators
    "$and",
    "$or",
    "$not",
    "$nor",  # Logical operators
    "$expr",
    "$jsonSchema",
    "$mod",
    "$regex",
    "$where",  # Other operators
    "$text",
    "$comment",
]  # Text search operators
SIMPLE_OPERATORS = [
    "$gt",
    "$gte",
    "$lt",
    "$lte",
    "$eq",
    "$ne",
    "$in",
    "$nin",
    "$exists",
    "$type",
    "$mod",
    "$regex",
    "$size",
    "$all",
    "$bitsAllClear",
    "$bitsAllSet",
    "$bitsAnyClear",
    "$bitsAnySet",
    "$box",
    "$center",
    "$centerSphere",
    "$geoIntersects",
    "$geometry",
    "$geoWithin",
    "$maxDistance",
    "$minDistance",
    "$near",
    "$nearSphere",
    "$polygon",
    "$text",
    "$comment",
]
COMPLEX_OPERATORS = [
    "$elemMatch",
    "$and",
    "$or",
    "$not",
    "$nor",
    "$expr",
    "$jsonSchema",
]


def analyze_query_pattern(log_line):
    query_type = "command"
    query = {}
    sort = {}
    msg = log_line.get("msg", "")
    if msg != "Slow query":
        return None
    attr = log_line.get("attr", {})
    op_type = attr.get("type", "")
    command = attr.get("command", {})
    if op_type == "update":
        # The real update command
        query_type = "update"
        query = command.get("q", {})
    elif "update" in command:
        # The update command can contain multiple updates
        query_type = "update.$cmd"
        query = command.get("updates", [])
    elif "aggregate" in command:
        query_type = "aggregate"
        query = command.get("pipeline", [])
        # This is not correct, but should cover 90% of cases
        # We only handle simple $match stage for now
        first_stage = query[0] if len(query) > 0 else {}
        if "$match" in first_stage:
            query = first_stage["$match"]
        # TODO: enumerate all stages to find out $sort stage.
    elif "find" in command:
        query_type = "find"
        query = command.get("filter", {})
        sort = command.get("sort", {})
    elif "getMore" in command:
        query_type = "getmore"
        query = attr.get("originatingCommand", {}).get("filter", {})
    elif "insert" in command:
        query_type = "insert"
        query = {}
    elif "delete" in command:
        query_type = "remove.$cmd"
        query = command.get("deletes", [])
    elif op_type == "remove":
        query_type = "remove"
        query = command.get("q", {})
    elif "findAndModify" in command:
        query_type = "findandmodify"
        query = command.get("query", {})
        sort = command.get("sort", {})

    if isinstance(query, list):
        # For list of queries, e.g., update.$cmd, remove.$cmd
        patterns = {}
        for q in query:
            q_pattern = query_to_pattern(q.get("q", {}))
            q_hash = json_hash(q_pattern, 4)
            patterns[q_hash] = q_pattern
        return {
            "type": query_type,
            "pattern": list(patterns.values()),
            "hash": list(patterns.keys()),
        }
    # For single query
    q_pattern = query_to_pattern(query)
    return {
        "type": query_type,
        "pattern": q_pattern,
        "sort": sort,
        "hash": json_hash({"query": q_pattern, "sort": sort}, 4),
    }


def query_to_pattern(query):
    shape = {}
    if isinstance(query, list):
        shape = [query_to_pattern(i) for i in query]
        # If all elements are 1, simplify to 1
        if all(i == 1 for i in shape):
            shape = 1
    elif isinstance(query, dict):
        for k, v in query.items():
            if k in COMPLEX_OPERATORS:
                shape[k] = query_to_pattern(v)
            else:
                shape[k] = _query_to_pattern(v)
    return shape


def _query_to_pattern(query):
    shape = {}
    if isinstance(query, list):
        shape = [_query_to_pattern(i) for i in query]
        # If all elements are 1, simplify to 1
        if all(i == 1 for i in shape):
            shape = 1
    elif isinstance(query, dict):
        for k, v in query.items():
            if k in COMPLEX_OPERATORS:
                shape[k] = query_to_pattern(v)
            elif k in SIMPLE_OPERATORS:
                shape[k] = 1
            else:
                shape = 1
    else:
        shape = 1
    return shape



"""Identify the top N slowest operations from the log entries."""

from bson import json_util
from x_ray.log_analysis.query_analyzer import analyze_query_pattern
from x_ray.log_analysis.log_items.base_item import BaseItem
from x_ray.utils import escape_markdown, format_json_md, json_hash


class TopSlowItem(BaseItem):
    """
    Identify the top N slowest operations from the log entries.
    """

    def __init__(self, output_folder: str, config):
        super().__init__(output_folder, config)
        self._top_n = config.get("top", 10)
        self.name = "Top Slow Operations"
        self.description = f"Identify the top `{self._top_n}` slowest operations from the log entries."
        self._cache = {}

    def analyze(self, log_line):
        log_id = log_line.get("id", "")
        if log_id != 51803:  # Slow query
            return
        attr = log_line.get("attr", {})
        ns = attr.get("ns", "")
        # Skip system namespaces
        if ns.startswith("admin.") or ns.startswith("local.") or ns.startswith("config."):
            return
        duration = attr.get("durationMillis", 0)
        has_sort = attr.get("hasSortStage", False)
        query_hash = attr.get("queryHash", "")
        n_returned = attr.get("nreturned", 0)
        keys_examined = attr.get("keysExamined", 0)
        docs_examined = attr.get("docsExamined", 0)
        plan_summary = attr.get("planSummary", "")
        query_pattern = analyze_query_pattern(log_line)
        if query_hash == "":
            # Some command doesn't have queryHash, e.g., getMore
            # If so, we generate one based on the query shape and sort
            query_hash = json_hash(query_pattern if query_pattern else {}, 4)
            # query_hash = query_pattern.get("hash", "N/A") if query_pattern else "N/A"
        slow_query = self._cache.get(query_hash, None)
        if slow_query is None:
            slow_query = {}
            self._cache[query_hash] = slow_query
        slow_query.update(
            {
                "query_hash": query_hash,
                "ns": ns,
                "query_pattern": query_pattern,
                "duration": slow_query.get("duration", 0) + duration,
                "n_returned": slow_query.get("n_returned", 0) + n_returned,
                "keys_examined": slow_query.get("keys_examined", 0) + keys_examined,
                "docs_examined": slow_query.get("docs_examined", 0) + docs_examined,
                "plan_summary": (plan_summary if "plan_summary" not in slow_query else slow_query["plan_summary"]),
                "has_sort": has_sort or slow_query.get("has_sort", False),
                "count": slow_query.get("count", 0) + 1,
                "sample": (log_line if "sample" not in slow_query else slow_query["sample"]),
            }
        )

    def finalize_analysis(self):
        self._cache = list(sorted(self._cache.values(), key=lambda item: item["count"], reverse=True)[: self._top_n])
        # self._cache = list(sorted(self._cache.values(), key=lambda item: item["duration"], reverse=True)[:self._top_n])
        super().finalize_analysis()

    def review_results_markdown(self, f):
        super().review_results_markdown(f)
        f.write('<div id="top_slow_positioner"></div>\n\n')
        f.write("|Query Hash|Op|Pattern|Details|Plan Summary|\n")
        f.write("|---|---|---|---|---|\n")
        # Total Duration (ms)|Count|Avg Duration (ms)|Scanned / Returned|ScannedObj / Returned|Has Sort
        i = 0
        with open(self._output_file, "r", encoding="utf-8") as data:
            for line in data:
                line_json = json_util.loads(line)
                query_hash = line_json.get("query_hash", "N/A")
                ns = line_json.get("ns", "N/A")
                query_pattern = line_json.get("query_pattern", {})
                op = query_pattern.get("type", "UNKNOWN")
                pattern = query_pattern.get("pattern", {})
                # query_hash = query_hash if query_hash != "" else "N/A"
                duration = line_json.get("duration", 0)
                count = line_json.get("count", 0)
                avg_duration = round(duration / count, 2) if count > 0 else 0
                n_returned = line_json.get("n_returned", 0)
                keys_examined = line_json.get("keys_examined", 0)
                docs_examined = line_json.get("docs_examined", 0)
                has_sort = "Yes" if line_json.get("has_sort", False) else "No"
                scanned_per_returned = round(keys_examined / n_returned, 2) if n_returned > 0 else keys_examined
                scannedobj_per_returned = round(docs_examined / n_returned, 2) if n_returned > 0 else docs_examined
                details = {
                    "Total Duration (ms)": duration,
                    "Count": count,
                    "Avg Duration (ms)": avg_duration,
                    "Targeting": scanned_per_returned,
                    "Targeting (Obj)": scannedobj_per_returned,
                    "Has Sort": has_sort,
                }
                plan_summary = line_json.get("plan_summary", "N/A")
                plan_summary = escape_markdown(plan_summary if plan_summary != "" else "N/A")
                cols = [
                    f"[{query_hash}](#{i})",
                    f"`{op}` on `{ns}`",
                    f"<pre>{format_json_md(pattern)}</pre>",
                    f"<pre>{format_json_md(details)}</pre>",
                    f"{plan_summary}",
                ]
                f.write(f"|{'|'.join(cols)}|\n")
                i += 1
        f.write("\n```json\n")
        f.write("// Click query hash to display sample query...\n")
        f.write("```\n")



from datetime import datetime, timezone
import random
import re
from pathlib import Path
import logging
import markdown
from bson import json_util
from x_ray.healthcheck.shared import to_json
from x_ray.utils import load_classes, bold, green, yellow, cyan, get_script_path, env

logger = logging.getLogger(__name__)
LOG_CLASSES = load_classes("x_ray.log_analysis.log_items")


class Framework:
    _logset_name = ""

    def __init__(self, file_path: str, config: dict):
        self._file_path = file_path
        self._config = config
        self._logger = logging.getLogger(__name__)
        self._items = []
        now = str(datetime.now(tz=timezone.utc))
        self._timestamp = re.sub(r"[:\- ]", "", now.split(".", maxsplit=1)[0])
        self._logger.debug(to_json(self._config))
        self._log_start = None
        self._log_end = None
        if env == "development":
            self._logger.info(yellow("Running in development mode."))

    def _get_output_folder(self, output_folder: str):
        if env == "development":
            batch_folder = output_folder
        else:
            batch_folder = f"{output_folder}{self._logset_name}-{self._timestamp}/"
        Path(batch_folder).mkdir(parents=True, exist_ok=True)
        return batch_folder

    def run_logs_analysis(self, logset_name: str, *args, **kwargs):
        self._logset_name = logset_name
        # Create output folder if it doesn't exist
        output_folder = kwargs.get("output_folder", "output/")
        batch_folder = self._get_output_folder(output_folder)
        # Dynamically load the log checkset based on the name
        logsets = self._config.get("logsets", {})
        if not logset_name in logsets:
            self._logger.warning(
                yellow(f"Log checkset '{logset_name}' not found in configuration. Using default logset.")
            )
            logset_name = "default"
        ls = logsets[logset_name]
        self._logger.info("Running log checkset: %s", bold(cyan(logset_name)))

        self._items = []
        for item_name in ls.get("items", []):
            item_cls = LOG_CLASSES.get(item_name)
            if not item_cls:
                self._logger.warning(yellow(f"Log item '{item_name}' not found. Skipping."))
                continue
            # The config for the item can be specified in the `item_config` section, under the item class name.
            item_config = self._config.get("item_config", {}).get(item_name, {})
            item = item_cls(batch_folder, item_config)
            self._items.append(item)
            self._logger.info("Log analyze item loaded: %s", bold(cyan(item_name)))
        log_file = self._file_path
        rate = self._config.get("sample_rate", 1.0)
        # Read the log file line by line and pass each line to the log items for analysis
        log_line = None
        with open(log_file, "r", encoding="utf-8", errors="ignore") as f:
            counter = 0
            for line in f:
                counter += 1
                if counter % 10000 == 0:
                    self._logger.info("%s lines ingested...", green(counter))
                # Sampling based on the rate. For dealing with large log files.
                if random.random() > rate:
                    continue
                try:
                    log_line = json_util.loads(line)
                    if self._log_start is None:
                        self._log_start = log_line.get("t", None)
                    for item in self._items:
                        try:
                            item.analyze(log_line)
                        except Exception as e:
                            self._logger.warning(yellow(f"Log analysis item '{item.name}' failed: {e}"))
                            continue
                except Exception:
                    self._logger.warning(yellow(f"Failed to parse log line as JSON: {line.strip()}"))
                    continue
        self._log_end = log_line.get("t", None) if log_line else None
        for item in self._items:
            try:
                item.finalize_analysis()
            except Exception as e:
                self._logger.warning(yellow(f"Log analysis item '{item.name}' finalize failed: {e}"))
                continue

    def output_results(self, output_folder: str = "output/", fmt: str = "html"):
        batch_folder = self._get_output_folder(output_folder)
        output_file = f"{batch_folder}report.md"
        template_file = get_script_path(f"templates/{self._config.get('template', 'log/full.html')}")
        self._logger.info("Saving results to: %s", green(output_file))

        with open(output_file, "w", encoding="utf-8") as f:
            f.write("# Log Analysis Report\n")
            f.write(f"Generated at: `{str(datetime.now(tz=timezone.utc))} UTC`\n\n")
            f.write(f"Log path: `{self._file_path}`\n\n")
            f.write(f"Log analysis period: `{self._log_start.isoformat()}` to `{self._log_end.isoformat()}`\n\n")
            f.write("Histogram chart instructions:\n\n")
            f.write("- **zoom in/out:** _mouse wheel or pinch_\n")
            f.write("- **pan:** _shift+drag_\n")
            f.write("- **select time frame:** _drag_\n\n")
            for item in self._items:
                try:
                    item.review_results_markdown(f)
                except Exception as e:
                    self._logger.warning(yellow(f"Failed to generate markdown for log item '{item.name}': {e}"))
                    continue

        if fmt == "html":
            html_file = f"{batch_folder}report.html"
            self._logger.info("Converting markdown to HTML: %s", green(html_file))
            with open(html_file, "w", encoding="utf-8") as f:
                with open(output_file, "r", encoding="utf-8") as md_file:
                    html_content = markdown.markdown(
                        md_file.read(),
                        extensions=["tables", "fenced_code", "toc", "md_in_html"],
                    )
                # Load the template file
                with open(template_file, "r", encoding="utf-8") as tf:
                    template_content = tf.read()
                    # Replace the placeholder with the generated HTML content
                final_html = template_content.replace("{{ content }}", html_content)
                f.write(final_html)



from x_ray.healthcheck.rules.base_rule import BaseRule
from x_ray.healthcheck.issues import ISSUE, create_issue
from x_ray.healthcheck.shared import MEMBER_STATE


class RSStatusRule(BaseRule):
    def apply(self, data: object, **kwargs) -> tuple:
        """Check the replica set status for any issues.

        Args:
            data (object): The result from `replSetGetStatus` command.
            extra_info (dict, optional): Extra information such as host. Defaults to None.
        Returns:
            tuple: (list of issues found, list of parsed data)
        """
        host = kwargs.get("extra_info", {}).get("host", "unknown")
        result = []
        # Find primary in members
        primary_member = next(iter(m for m in data["members"] if m["state"] == 1), None)

        no_primary = False
        if not primary_member:
            no_primary = True
            issue = create_issue(ISSUE.NO_PRIMARY, host=host)
            result.append(issue)

        # Check member states
        max_delay = self._thresholds.get("replication_lag_seconds", 60)
        set_name = data.get("set", "Unknown Set")
        for member in data["members"]:
            # Check problematic states
            state = member["state"]
            host = member["name"]

            if state in [3, 6, 8, 9, 10]:
                issue = create_issue(
                    ISSUE.UNHEALTHY_MEMBER,
                    host=host,
                    params={"set_name": set_name, "host": host, "state": MEMBER_STATE[state]},
                )
                result.append(issue)
            elif state in [0, 5]:
                issue = create_issue(
                    ISSUE.INITIALIZING_MEMBER,
                    host=host,
                    params={"set_name": set_name, "host": host, "state": MEMBER_STATE[state]},
                )
                result.append(issue)

            # Check replication lag
            if state == 2 and not no_primary:  # SECONDARY
                p_time = primary_member["optime"]["ts"]
                s_time = member["optime"]["ts"]
                lag = p_time.time - s_time.time
                if lag >= max_delay:
                    issue = create_issue(
                        ISSUE.DELAYED_MEMBER, host=host, params={"set_name": set_name, "host": host, "lag": lag}
                    )
                    result.append(issue)
        return result, data



from x_ray.healthcheck.rules.base_rule import BaseRule
from x_ray.healthcheck.issues import ISSUE, create_issue


class ConnectionsRule(BaseRule):
    def __init__(self, thresholds=None):
        super().__init__(thresholds)
        self._used_connection_ratio = thresholds.get("used_connection_ratio", 0.8)

    def apply(self, data: object, **kwargs) -> tuple:
        """Check the connections usage for any issues.

        Args:
            data (object): The result from `serverStatus` command.
            extra_info (dict, optional): Extra information such as host. Defaults to None.
        Returns:
            tuple: (list of issues found, list of parsed data)
        """
        host = kwargs.get("extra_info", {}).get("host", "unknown")
        test_result = []
        connections = data.get("connections", {})
        available = connections.get("available", 0)
        current = connections.get("current", 0)
        total = available + current
        if current / total > self._used_connection_ratio:
            issue = create_issue(
                ISSUE.HIGH_CONNECTION_USAGE_RATIO,
                host=host,
                params={
                    "current": current,
                    "total": total,
                    "used_connection_ratio": self._used_connection_ratio * 100,
                },
            )
            test_result.append(issue)

        return test_result, connections



"""Build Info Check Item Module. Used to check MongoDB server build information."""

from x_ray.healthcheck.check_items.base_item import BaseItem
from x_ray.healthcheck.rules.version_eol_rule import VersionEOLRule
from x_ray.healthcheck.shared import MAX_MONGOS_PING_LATENCY, discover_nodes, enum_all_nodes, enum_result_items
from x_ray.utils import yellow
from x_ray.version import Version


class BuildInfoItem(BaseItem):
    def __init__(self, output_folder: str, config: dict = None):
        super().__init__(output_folder, config)
        self._name = "Build Information"
        self._description = "Collects & review server build information.\n\n"
        self._description += "- Whether the server is running a supported version.\n"
        self._version_eol_rule = VersionEOLRule(config)

    def test(self, *args, **kwargs):
        client = kwargs.get("client")
        parsed_uri = kwargs.get("parsed_uri")
        nodes = discover_nodes(client, parsed_uri)

        def func_node(set_name, node, **kwargs):
            host = node["host"]
            if "pingLatencySec" in node and node["pingLatencySec"] > MAX_MONGOS_PING_LATENCY:
                self._logger.warning(
                    yellow(
                        f"Skip {host} because it has been irresponsive for {node['pingLatencySec'] / 60:.2f} minutes."
                    )
                )
                return None, None
            client = node["client"]
            raw_result = client.admin.command("buildInfo")
            test_result, _ = self._version_eol_rule.apply(raw_result, extra_info={"host": host})
            running_version = Version(raw_result.get("versionArray", None))
            node["version"] = running_version
            self.append_test_results(test_result)

            return test_result, raw_result

        raw_result = enum_all_nodes(
            nodes,
            func_mongos_member=func_node,
            func_rs_member=func_node,
            func_shard_member=func_node,
            func_config_member=func_node,
        )

        self.captured_sample = raw_result

    @property
    def review_result(self):
        result = self.captured_sample
        table = {
            "type": "table",
            "caption": "Server Build Information",
            "columns": [
                {"name": "Component", "type": "string"},
                {"name": "Host", "type": "string"},
                {"name": "Version", "type": "string"},
                {"name": "OpenSSL", "type": "string"},
                {"name": "Target Arch", "type": "string"},
                {"name": "Target OS", "type": "string"},
            ],
            "rows": [],
        }
        versions = {}

        def func_node(name, node, **kwargs):
            raw_result = node.get("rawResult", {})
            host = node["host"]
            if raw_result is None:
                table["rows"].append([name, host, "n/a", "n/a", "n/a", "n/a"])
                versions["n/a"] = versions.get("n/a", 0) + 1
                return
            build_env = raw_result.get("buildEnvironment", {})
            v = raw_result.get("version", "")
            versions[v] = versions.get(v, 0) + 1
            table["rows"].append(
                [
                    name,
                    host,
                    v,
                    raw_result.get("openssl", {}).get("running", ""),
                    build_env.get("target_arch", ""),
                    build_env.get("target_os", ""),
                ]
            )

        enum_result_items(
            result,
            func_mongos_member=func_node,
            func_rs_member=func_node,
            func_shard_member=func_node,
            func_config_member=func_node,
        )
        version_pie = {"type": "chart", "data": versions}

        return {"name": self.name, "description": self.description, "data": [table, version_pie]}



from x_ray.healthcheck.rules.base_rule import BaseRule
from x_ray.healthcheck.issues import ISSUE, create_issue


class ShardKeyRule(BaseRule):
    def apply(self, data: object, **kwargs) -> tuple:
        """Check shard key configurations for issues.

        Args:
            data (object): One document from `config.collections`.
            extra_info (dict, optional): Extra information such as host.
        Returns:
            tuple: (list of issues found, list of parsed data)
        """
        test_results = []
        ns = data["_id"]
        key = data["key"]
        v = key.get("_id", None)
        if v in [-1, 1] and len(key.keys()) == 1:
            issue = create_issue(
                ISSUE.IMPROPER_SHARD_KEY,
                host="cluster",
                params={
                    "ns": ns,
                    "shard_key": f"{{_id: {v}}}",
                },
            )
            test_results.append(issue)
        return test_results, data



from x_ray.healthcheck.rules.base_rule import BaseRule
from x_ray.healthcheck.issues import ISSUE, create_issue


class RSConfigRule(BaseRule):
    def apply(self, data: object, **kwargs) -> tuple:
        """Check the replica set configuration for any issues.

        Args:
            data (object): The replica set configuration data.
            extra_info (dict, optional): Extra information such as host. Defaults to None.
        Returns:
            tuple: (list of issues found, list of parsed data)
        """
        result = []
        set_name = data["config"]["_id"]
        # Check number of voting members
        voting_members = sum(1 for member in data["config"]["members"] if member.get("votes", 0) > 0)
        if voting_members < 3:
            issue = create_issue(
                ISSUE.INSUFFICIENT_VOTING_MEMBERS,
                host="cluster",
                params={"set_name": set_name, "voting_members": voting_members},
            )
            result.append(issue)
        if voting_members % 2 == 0:
            issue = create_issue(
                ISSUE.EVEN_VOTING_MEMBERS,
                host="cluster",
                params={"set_name": set_name},
            )
            result.append(issue)

        for member in data["config"]["members"]:
            delay = member.get("secondaryDelaySecs", member.get("slaveDelay", 0))
            if delay > 0:
                if member.get("votes", 0) > 0:
                    issue = create_issue(
                        ISSUE.DELAYED_VOTING_MEMBER,
                        host=member["host"],
                        params={"set_name": set_name, "host": member["host"]},
                    )
                    result.append(issue)
                if member.get("priority", 0) > 0:
                    issue = create_issue(
                        ISSUE.DELAYED_ELECTABLE_MEMBER,
                        host=member["host"],
                        params={"set_name": set_name, "host": member["host"]},
                    )
                    result.append(issue)
                if not member.get("hidden", False):
                    issue = create_issue(
                        ISSUE.DELAYED_NON_HIDDEN_MEMBER,
                        host=member["host"],
                        params={"set_name": set_name, "host": member["host"]},
                    )
                    result.append(issue)

                issue = create_issue(
                    ISSUE.DELAYED_SECONDARY_MEMBER,
                    host=member["host"],
                    params={"set_name": set_name, "host": member["host"]},
                )
                result.append(issue)
            if member.get("arbiterOnly", False):
                issue = create_issue(
                    ISSUE.ARBITER_MEMBER,
                    host=member["host"],
                    params={"set_name": set_name, "host": member["host"]},
                )
                result.append(issue)
        return result, data



"""
Analyze connection rates from log entries.
"""

from datetime import datetime
import math
from x_ray.log_analysis.log_items.base_item import BaseItem


class ConnectionRateItem(BaseItem):
    def __init__(self, output_folder: str, config):
        super().__init__(output_folder, config, show_reset=True)
        self._cache = None
        self.name = "Connection Rate"
        self.description = "Analyse the rate of connections created and ended over a specified time window."

    def analyze(self, log_line):
        log_id = log_line.get("id", "")
        if log_id not in [22943, 22944]:  # Connection accepted/ended
            return
        if self._cache is None:
            self._cache = {}
        counter = "created" if log_id == 22943 else "ended"
        time = log_line.get("t")
        ts = math.floor(time.timestamp())
        time_min = datetime.fromtimestamp(ts - (ts % 60))

        if self._cache.get("time", None) != time_min:
            if self._cache != {}:
                self._write_output()
            self._cache = {
                "time": time_min,
                "created": 0,
                "ended": 0,
                "total": 0,
                "byIp": {},
            }
        attr = log_line.get("attr", {})
        conn_count = attr.get("connectionCount", 1)
        ip = attr["remote"].split(":")[0] if "remote" in attr else "unknown"
        self._cache[counter] += 1
        self._cache["total"] = conn_count
        if ip not in self._cache["byIp"]:
            self._cache["byIp"][ip] = {"created": 0, "ended": 0}
        self._cache["byIp"][ip][counter] += 1

    def review_results_markdown(self, f):
        super().review_results_markdown(f)
        f.write(f'<canvas id="canvas_{self.__class__.__name__}" width="400" height="200"></canvas>\n')
        f.write(f'<canvas id="canvas_{self.__class__.__name__}_byip" width="400" height="200"></canvas>\n')



from datetime import datetime
from x_ray.utils import to_ejson


def to_json(obj, indent=None):
    cls_maps = [{"class": datetime, "func": lambda o: o.isoformat()}]
    return to_ejson(obj, indent=indent, cls_maps=cls_maps)



from x_ray.healthcheck.check_items.base_item import BaseItem
from x_ray.healthcheck.rules.host_info_rule import HostInfoRule
from x_ray.healthcheck.rules.numa_rule import NumaRule
from x_ray.healthcheck.shared import MAX_MONGOS_PING_LATENCY, discover_nodes, enum_all_nodes, enum_result_items
from x_ray.utils import yellow, format_size


class HostInfoItem(BaseItem):
    def __init__(self, output_folder, config=None):
        super().__init__(output_folder, config)
        self._name = "Host Information"
        self._description = "Collects and reviews host hardware and OS information.  \n"
        self._description += "- Whether the hosts are using the same hardware.\n\n"
        self._description += "- Whether NUMA is enabled on the hosts.\n"
        self._host_info_rule = HostInfoRule(config)
        self._numa_rule = NumaRule(config)

    def test(self, *args, **kwargs):
        """
        Main test method to gather host information.
        """
        client = kwargs.get("client")
        parsed_uri = kwargs.get("parsed_uri")
        nodes = discover_nodes(client, parsed_uri)

        host_infos = {}

        def func_single(name, node, **kwargs):
            client = node["client"]
            version = node.get("version", None)
            if "pingLatencySec" in node and node["pingLatencySec"] > MAX_MONGOS_PING_LATENCY:
                self._logger.warning(
                    yellow(
                        f"Skip {node['host']} because it has been irresponsive for {node['pingLatencySec'] / 60:.2f} minutes."
                    )
                )
                return None, None
            host_info = client.admin.command("hostInfo")
            test_result, _ = self._numa_rule.apply(host_info, extra_info={"host": node["host"], "version": version})
            if name not in host_infos:
                host_infos[name] = []
            host_infos[name].append(host_info)
            return test_result, host_info

        result = enum_all_nodes(
            nodes,
            func_rs_member=func_single,
            func_mongos_member=func_single,
            func_shard_member=func_single,
            func_config_member=func_single,
        )
        for set_name, info in host_infos.items():
            test_result, _ = self._host_info_rule.apply(info, extra_info={"set_name": set_name})
            cluster_map = result["map"]
            if set_name not in cluster_map:
                cluster = cluster_map["config"]
            else:
                cluster = cluster_map[set_name]
            cluster["testResult"] = test_result

        self.captured_sample = result

    @property
    def review_result(self):
        """
        Review the gathered host information.
        """
        result = self.captured_sample
        data = []

        def func_component(name, node, **kwargs):
            members = node["members"]
            table = {
                "type": "table",
                "caption": f"Hardware & OS Information - `{name}`",
                "columns": [
                    {"name": "Host", "type": "string"},
                    {"name": "CPU Family", "type": "string"},
                    {"name": "CPU Cores", "type": "string"},
                    {"name": "Memory", "type": "string"},
                    {"name": "OS", "type": "string"},
                    {"name": "NUMA", "type": "boolean"},
                ],
                "rows": [],
            }
            data.append(table)
            for m in members:
                info = m["rawResult"]
                if info is None:
                    table["rows"].append([m["host"], "n/a", "n/a", "n/a", "n/a", "n/a"])
                    continue
                system = info["system"]
                os = info["os"]
                extra = info["extra"]
                if "extra" in extra:
                    # Compatibility for MongoDB 6.0
                    extra = extra["extra"]
                table["rows"].append(
                    [
                        m["host"],
                        f"{extra.get('cpuString', '(Unknown CPU)')} ({system['cpuArch']}) {extra.get('cpuFrequencyMHz', 'n/a')} MHz",
                        f"{system['numCores']}c",
                        format_size(system["memSizeMB"] * 1024**2),
                        f"{os['name']} {os['version']}",
                        system["numaEnabled"],
                    ]
                )

        enum_result_items(
            result,
            func_rs_cluster=func_component,
            func_all_mongos=func_component,
            func_shard=func_component,
            func_config=func_component,
        )
        return {"name": self.name, "description": self.description, "data": data}



"""Utility functions and classes for the X-Ray project."""

from importlib import import_module
from importlib.resources import files
import os
import json
import logging
import pkgutil
import re
import hashlib
import sys
from enum import Enum
from pathlib import Path
from bson import json_util
from x_ray.version import Version

levels = logging._nameToLevel
level = os.getenv("LOG_LEVEL", "INFO")
env = os.getenv("ENV", "production")
ai_key = os.getenv("OPENAI_API_KEY", "")
if level not in levels:
    level = "INFO"
log_level = levels[level]
logging.basicConfig(level=log_level)
logger = logging.getLogger(__name__)
logger.debug("Using log level: %s", level)

BUILDIN_CONFIG_PATH = "config.json"


# The script can be started from other working folder. E.g. Invoked by a cron job.
# This function gives you the base path of the script.
# If `filename` is provided then the path include the file. Otherwise it's the folder.
def get_script_path(filename=None):
    # Check if running in a PyInstaller bundle
    if getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS"):
        # Running in a PyInstaller bundle
        base_path = sys._MEIPASS
        if filename is None:
            return base_path
        # In PyInstaller, resources are under x_ray/ subdirectory
        return str(Path(base_path) / "x_ray" / filename)

    # Running in normal Python environment - use importlib.resources
    if filename is None:
        # Return package directory
        return str(files("x_ray"))

    # Return specific file path
    resource_path = files("x_ray") / filename
    return str(resource_path)


def _load_config():
    # Use a mutable container to hold cached config to avoid nonlocal assignment warnings.
    config = None

    def func(config_path):
        nonlocal config
        if config is None:
            if config_path is None:
                resource = files("x_ray") / BUILDIN_CONFIG_PATH
                config = json.loads(resource.read_text(encoding="utf-8"))
                logger.info("Loaded built-in config: %s", BUILDIN_CONFIG_PATH)
            elif os.path.isfile(config_path):
                # Try to load from the path provided by the user
                with open(config_path, "r", encoding="utf-8") as f:
                    config = json.load(f)
                logger.info("Loaded config from user-provided path: %s", config_path)
            else:
                # If all fails, raise an error
                logger.error("Failed to load config file: %s", config_path)
                raise FileNotFoundError(f"Could not find config file: {config_path}")

        return config

    return func


load_config = _load_config()


MAX_CONTENT_WORDS = 3
MORE_CONTENT = "..."


def truncate_content(
    content: str,
    delimiter="[ \t]",
    max_words=MAX_CONTENT_WORDS,
    more_content=MORE_CONTENT,
) -> str:
    content = content.strip()
    truncated = re.split(delimiter, content)
    if len(truncated) <= max_words:
        return content
    return " ".join(truncated[:max_words]) + f" {more_content}"


def tooltip_html(full, truncated) -> str:
    html = f'<span class="tooltip" data-tip="{full}">{truncated}</span>'
    return html


def load_classes(package_name="x_ray.log_analysis.log_items"):
    class_map = {}
    package = import_module(package_name)
    for _, module_name, _ in pkgutil.iter_modules(package.__path__):
        module = import_module(f"{package_name}.{module_name}")
        for attr in dir(module):
            obj = getattr(module, attr)
            if isinstance(obj, type):
                class_map[attr] = obj
    logger.debug("Loaded getMongoData analysis classes: %s", list(class_map.keys()))
    return class_map


def format_size(size_bytes, decimal=2):
    """
    Format the size in bytes to a human-readable string.

    Args:
        bytes (int): The size in bytes.
        decimal (int): The number of decimal places to include.

    Returns:
        str: The formatted size string.
    """
    for unit in ["B", "KB", "MB", "GB", "TB"]:
        if size_bytes < 1024:
            return f"{size_bytes:.{decimal}f} {unit}"
        size_bytes /= 1024
    return f"{size_bytes:.{decimal}f} PB"


def escape_markdown(text):
    """
    Escape markdown special characters.
    """
    ESCAPE_MAP = {
        "_": "\\_",
        "*": "\\*",
        "`": "\\`",
        "|": "\\|",
        "<": "&lt;",
        ">": "&gt;",
    }
    if not isinstance(text, str):
        text = str(text)
    # Escape underscores, asterisks, backticks, and other special characters
    for key, value in ESCAPE_MAP.items():
        text = text.replace(key, value)
    return text


def format_json_md(json_data, **kwargs):
    """
    Format JSON data as a markdown code block.
    If indent is None or 0, returns a compressed JSON string without line breaks.
    """
    indent = kwargs.get("indent", 2)
    if indent is None or indent == 0:
        kwargs["separators"] = (",", ": ")
        kwargs["indent"] = None
        json_str = to_ejson(json_data, **kwargs)
    else:
        json_str = to_ejson(json_data, **kwargs).replace(" ", "&nbsp;").replace("\n", "<br>")
    return json_str


def to_ejson(obj, **kwargs):
    indent = kwargs.pop("indent", 2)
    separators = kwargs.pop("separators", None)
    cls_maps = [
        {"class": Enum, "func": lambda o: o.name},
        {"class": Version, "func": str},
    ]
    cls_maps.extend(kwargs.pop("cls_maps", []))

    def custom_serializer(o):
        for cls_map in cls_maps:
            cls = cls_map.get("class", None)
            func = cls_map.get("func", None)
            if cls and func and isinstance(o, cls):
                return func(o)
        return json_util.default(o)

    # Must use json.dumps because bson.json_util.dumps has its own serializer behavior,
    # and won't always call our custom_serializer.
    # It only calls when the object is not serializable by default.
    return json.dumps(obj, indent=indent, separators=separators, default=custom_serializer)


def json_hash(data, digest_size=8):
    json_str = to_ejson(data, indent=None)
    h = hashlib.blake2b(json_str.encode("utf-8"), digest_size=digest_size)
    return h.digest().hex().upper()


def color_code(code):
    return f"\x1b[{code}m"


def colorize(code: int, s: str) -> str:
    return f"{color_code(code)}{str(s).replace(color_code(0), color_code(code))}{color_code(0)}"


def green(s: str) -> str:
    return colorize(32, s)


def yellow(s: str) -> str:
    return colorize(33, s)


def red(s: str) -> str:
    return colorize(31, s)


def cyan(s: str) -> str:
    return colorize(36, s)


def magenta(s: str) -> str:
    return colorize(35, s)


def bold(s: str) -> str:
    return colorize(1, s)


def dim(s: str) -> str:
    return colorize(2, s)


def italic(s: str) -> str:
    return colorize(3, s)


def underline(s: str) -> str:
    return colorize(4, s)


def blink(s: str) -> str:
    return colorize(5, s)


def reverse(s: str) -> str:
    return colorize(7, s)


def invisible(s: str) -> str:
    return colorize(8, s)



from x_ray.healthcheck.rules.base_rule import BaseRule
from x_ray.healthcheck.issues import ISSUE, create_issue
from x_ray.version import Version


class VersionEOLRule(BaseRule):
    def apply(self, data: object, **kwargs) -> tuple:
        """Check if the given build info represents a version that is end-of-life (EOL).

        Args:
            data (object): The result from `buildInfo` command.
            extra_info (dict): Additional information such as host.
        Returns:
            tuple: (list of issues found, list of parsed data)
        """
        host = kwargs.get("extra_info", {}).get("host", "unknown")
        test_results = []
        version = Version(data.get("versionArray", [0, 0, 0]))
        eol_version = Version(self._thresholds.get("eol_version", [6, 3, 0]))
        # Check if the version is below EOL version
        if version < eol_version:
            issue = create_issue(
                ISSUE.EOL_VERSION_USED, host=host, params={"version": version, "eol_version": eol_version}
            )
            test_results.append(issue)
        # Check if rapid releases are being used
        if 5 <= version.major_version <= 7 and version.minor_version != 0:
            issue = create_issue(ISSUE.RAPID_RELEASE_VERSION_USED, host=host, params={"version": version})
            test_results.append(issue)
        # Check if development releases are being used
        if version.major_version >= 8 or version.major_version <= 4:
            if version.minor_version % 2 != 0:
                issue = create_issue(ISSUE.DEVELOPMENT_RELEASE_VERSION_USED, host=host, params={"version": version})
                test_results.append(issue)
        return test_results, data



from x_ray.version import Version


def test_version_comparison():
    v1 = Version.parse("1.2.3")
    v2 = Version([1, 2, 4])
    v3 = Version.parse("1.3.0")
    v4 = Version.parse("2.0.0")
    v5 = Version.parse("1.2.3")

    assert v1 < v2
    assert v2 < v3
    assert v3 < v4
    assert v1 == v5
    assert v4 > v1
    assert v3 >= v2
    assert v2 <= v4
    assert v1 < "1.2.4"
    assert v3 > [1, 2, 9]



from x_ray.healthcheck.rules.base_rule import BaseRule
from x_ray.healthcheck.issues import ISSUE, create_issue


class DataSizeRule(BaseRule):
    def __init__(self, thresholds=None):
        super().__init__(thresholds)
        self._collection_size_gb = self._thresholds.get("collection_size_gb", 2048) * 1024**3
        self._obj_size_bytes = self._thresholds.get("obj_size_kb", 32) * 1024
        self._index_size_ratio = self._thresholds.get("index_size_ratio", 0.2)

    def apply(self, data: object, **kwargs) -> tuple:
        """Check the data size for any issues.

        Args:
            data (object): The data size status data.
            extra_info (dict, optional): Extra information such as host. Defaults to None.
        Returns:
            tuple: (list of issues found, list of parsed data)
        """
        host = kwargs.get("extra_info", {}).get("host", "unknown")
        test_result = []
        storage_stats = data.get("storageStats", {})
        del storage_stats["indexDetails"]
        del storage_stats["wiredTiger"]
        # Check for large collection size
        if storage_stats.get("size", 0) > self._collection_size_gb:
            issue = create_issue(
                ISSUE.COLLECTION_TOO_LARGE,
                host=host,
                params={
                    "ns": data.get("ns", ""),
                    "size_gb": storage_stats.get("size", 0) / 1024**3,
                    "collection_size_gb": self._collection_size_gb / 1024**3,
                },
            )
            test_result.append(issue)
        # Check for average object size
        if storage_stats.get("avgObjSize", 0) > self._obj_size_bytes:
            issue = create_issue(
                ISSUE.AVG_OBJECT_SIZE_TOO_LARGE,
                host=host,
                params={
                    "ns": data.get("ns", ""),
                    "avg_obj_size_kb": storage_stats.get("avgObjSize", 0) / 1024,
                    "obj_size_kb": self._obj_size_bytes / 1024,
                },
            )
            test_result.append(issue)

        return test_result, data



from x_ray.healthcheck.rules.base_rule import BaseRule
from x_ray.healthcheck.issues import ISSUE, create_issue


class SecurityRule(BaseRule):
    def apply(self, data: object, **kwargs) -> tuple:
        """Check the security settings for any issues.

        Args:
            data (object): The getCmdLineOpts data.
            extra_info (dict, optional): Extra information such as host. Defaults to None.
        Returns:
            tuple: (list of issues found, list of parsed data)
        """
        host = kwargs.get("extra_info", {}).get("host", "unknown")
        test_result = []
        parsed = data.get("parsed", {})
        security_settings = parsed.get("security", {})
        net = parsed.get("net", {})
        audit_log = parsed.get("auditLog", {})
        authorization = security_settings.get("authorization", None)
        redact_logs = security_settings.get("redactClientLogData", None)
        bind_ip = net.get("bindIp", "127.0.0.1")
        port = net.get("port", None)
        tls_enabled = net.get("tls", {}).get("mode", None)
        audit = "enabled" if audit_log.get("destination", None) is not None else "disabled"
        ear_enabled = security_settings.get("enableEncryption", False)
        ear_keyfile = security_settings.get("encryptionKeyFile", None)
        if authorization != "enabled":
            issue = create_issue(ISSUE.AUTHORIZATION_DISABLED, host=host)
            test_result.append(issue)
        if not redact_logs:
            issue = create_issue(ISSUE.LOG_REDACTION_DISABLED, host=host)
            test_result.append(issue)
        if tls_enabled is None:
            issue = create_issue(ISSUE.TLS_DISABLED, host=host)
            test_result.append(issue)
        elif tls_enabled != "requireTLS":
            issue = create_issue(ISSUE.OPTIONAL_TLS, host=host, params={"tls_mode": tls_enabled})
            test_result.append(issue)
        if bind_ip == "0.0.0.0":
            issue = create_issue(ISSUE.OPEN_BIND_IP, host=host)
            test_result.append(issue)
        if port == 27017:
            issue = create_issue(ISSUE.DEFAULT_PORT_USED, host=host)
            test_result.append(issue)
        if audit == "disabled":
            issue = create_issue(ISSUE.AUDITING_DISABLED, host=host)
            test_result.append(issue)
        if not ear_enabled:
            issue = create_issue(ISSUE.ENCRYPTION_AT_REST_DISABLED, host=host)
            test_result.append(issue)
        if ear_keyfile is not None:
            issue = create_issue(ISSUE.ENCRYPTION_AT_REST_USING_KEYFILE, host=host)
            test_result.append(issue)

        return test_result, data



from x_ray.log_analysis.log_items.base_item import BaseItem


class StateTraceItem(BaseItem):
    LOG_IDS = [
        4615611,  # starting (for rs members)
        20721,  # starting (for shard members)
        20722,  # new state
        21392,  # new configuration applied
        21215,  # new state
        21216,  # new state
        21358,  # state transition
        4615660,  # priority takeover
    ]

    def __init__(self, output_folder, config):
        super().__init__(output_folder, config, show_reset=True)
        self.name = "Member State Trace"
        self.description = "Visualize member state trace logs to understand system state changes over time."
        self._cache = None
        self._myself = "self"
        self._last_log = None

    def analyze(self, log_line):
        super().analyze(log_line)
        log_id = log_line.get("id", "")
        self._last_log = log_line
        if log_id not in self.LOG_IDS:
            return
        if self._cache is None:
            self._cache = {}
        msg = log_line.get("msg", "")
        if log_id in [4615611, 20721]:
            # Identify myself
            host = log_line.get("attr", {}).get("host", "unknown")
            port = log_line.get("attr", {}).get("port", "unknown")
            self._myself = f"{host}:{port}"
        if log_id == 20722:
            # New state
            host = self._myself
            new_state = log_line.get("attr", {}).get("memberState", "unknown")
            if host not in self._cache:
                self._cache[host] = []
            self._cache[host].append(
                {
                    "id": log_id,
                    "host": host,
                    "timestamp": log_line.get("t", ""),
                    "event": "NewMemberState",
                    "details": {"new_state": new_state, "msg": "Initial State"},
                }
            )
        if log_id == 21392:
            # New configuration applied
            host = self._myself
            config = log_line.get("attr", {}).get("config", {})
            if host not in self._cache:
                self._cache[host] = []
            self._cache[host].append(
                {
                    "id": log_id,
                    "host": host,
                    "timestamp": log_line.get("t", ""),
                    "event": "NewConfig",
                    "details": {"msg": msg, "config": config},
                }
            )
        if log_id == 21215:
            # New state for others
            host = log_line.get("attr", {}).get("hostAndPort", {})
            new_state = log_line.get("attr", {}).get("newState", "unknown")
            if host not in self._cache:
                self._cache[host] = []
            self._cache[host].append(
                {
                    "id": log_id,
                    "host": host,
                    "timestamp": log_line.get("t", ""),
                    "event": "NewMemberState",
                    "details": {"new_state": new_state, "msg": msg},
                }
            )
        if log_id == 21216:
            # New state
            host = log_line.get("attr", {}).get("hostAndPort", {})
            # match new state from msg
            new_state = msg.split(" ")[-1] if msg else "UNKNOWN"
            # normalize RS_DOWN to DOWN because according to the document DOWN is the correct state name.
            if new_state == "RS_DOWN":
                new_state = "DOWN"
            if host not in self._cache:
                self._cache[host] = []
            self._cache[host].append(
                {
                    "id": log_id,
                    "host": host,
                    "timestamp": log_line.get("t", ""),
                    "event": "NewMemberState",
                    "details": {"new_state": new_state, "msg": msg},
                }
            )
        if log_id == 21358:
            # State transition
            host = self._myself
            old_state = log_line.get("attr", {}).get("oldState", "unknown")
            new_state = log_line.get("attr", {}).get("newState", "unknown")
            if host not in self._cache:
                self._cache[host] = []
            self._cache[host].append(
                {
                    "id": log_id,
                    "host": host,
                    "timestamp": log_line.get("t", ""),
                    "event": "StateTransition",
                    "details": {
                        "msg": msg,
                        "old_state": old_state,
                        "new_state": new_state,
                    },
                }
            )
        if log_id == 4615660:
            # Priority takeover
            host = self._myself
            msg = log_line.get("msg", "")
            if host not in self._cache:
                self._cache[host] = []
            self._cache[host].append(
                {
                    "id": log_id,
                    "host": host,
                    "timestamp": log_line.get("t", ""),
                    "event": "PriorityTakeover",
                    "details": {"msg": msg},
                }
            )

    def finalize_analysis(self):
        if self._last_log:
            # Because we are using line chart to describe the state changes,
            # we need to add a final point to indicate the last known state.
            # Otherwise, the chart will end abruptly at the last event.
            for host, events in self._cache.items():
                events.append(
                    {
                        "id": -1,
                        "host": host,
                        "timestamp": (self._last_log.get("t", "") if self._last_log else ""),
                        "event": "LogEnd",
                        "details": {"msg": "End of log"},
                    }
                )
        super().finalize_analysis()

    def review_results_markdown(self, f):
        super().review_results_markdown(f)
        f.write(f'<canvas id="canvas_{self.__class__.__name__}" height="300" width="400"></canvas>\n')



from x_ray.healthcheck.rules.base_rule import BaseRule
from x_ray.healthcheck.issues import ISSUE, create_issue


class QueryTargetingRule(BaseRule):
    def __init__(self, thresholds=None):
        super().__init__(thresholds)
        self._max_query_targeting = thresholds.get("query_targeting", 0.1)
        self._max_query_targeting_obj = thresholds.get("query_targeting_obj", 0.1)

    def apply(self, data: object, **kwargs) -> tuple:
        """Check query targeting efficiency.

        Args:
            data (object): The result from `serverStatus` command.
            extra_info (dict, optional): Extra information such as host. Defaults to None.
        Returns:
            tuple: (list of issues found, list of parsed data)
        """
        host = kwargs.get("extra_info", {}).get("host", "unknown")
        test_result = []
        query_executor = data["metrics"].get("queryExecutor", {})
        document = data["metrics"].get("document", {})
        scanned_returned = (
            (query_executor["scanned"] / document["returned"])
            if document["returned"] > 0
            else query_executor["scanned"]
        )
        scanned_obj_returned = (
            (query_executor["scannedObjects"] / document["returned"])
            if document["returned"] > 0
            else query_executor["scannedObjects"]
        )
        if scanned_returned > self._max_query_targeting:
            issue = create_issue(
                ISSUE.POOR_QUERY_TARGETING_KEYS,
                host=host,
                params={
                    "scanned_returned": scanned_returned,
                    "query_targeting": self._max_query_targeting,
                },
            )
            test_result.append(issue)
        if scanned_obj_returned > self._max_query_targeting_obj:
            issue = create_issue(
                ISSUE.POOR_QUERY_TARGETING_OBJECTS,
                host=host,
                params={
                    "scanned_obj_returned": scanned_obj_returned,
                    "query_targeting_obj": self._max_query_targeting_obj,
                },
            )
            test_result.append(issue)

        return test_result, {"scanned/returned": scanned_returned, "scanned_obj/returned": scanned_obj_returned}
